"""
    Copyright 2015-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import math
from .. import eggwrapper as _eggwrapper
from .. import _interface
from .. import _freq
from .. import _site
from .. io import _vcf

class ComputeStats(object):
    """ComputeStats(...)

    Customizable and efficient analysis of diversity
    data. This class is designed to minimize redundancy of underlying analyses
    so it is best to compute as many statistics as possible with a
    single instance. It also takes advantage of the object reuse policy,
    improving the efficiency of analyses when several (and especially
    many) datasets are examined in a row by the same instance of
    :class:`~.ComputeStats`.

    The constructor takes arguments that are automatically passed to the
    :meth:`configure` method.

    Statistics to compute are set using the method :meth:`~.ComputeStats.add_stats`
    which allows specifying several statistics at once and which can
    also be called several times to add more statistics to compute.
    """

    def list_stats(self):
        """
        List of available statistics. Returns a list of tuples giving, for each
        available statistic, its code and a short description.
        """
        return [(k, v[0]) for (k, v) in self._stats.items()]

    def _get_genosite_from_store(self):
        if len(self._genosite_store): return self._genosite_store.pop()
        else: return _eggwrapper.Genotypes()

    def __init__(self, *args, **kwargs):
        self._sd = _eggwrapper.SiteDiversity()
        self._sd2 = _eggwrapper.SiteDiversity()
        self._as = _eggwrapper.AlleleStatus()
        self._d1 = _eggwrapper.Diversity1()
        self._d2 = _eggwrapper.Diversity2()
        self._h = _eggwrapper.Haplotypes()
        self._rd = _eggwrapper.Rd()
        self._ld = _eggwrapper.MatrixLD()
        self._cv = _eggwrapper.ComputeV()
        self._frq = _eggwrapper.FreqHolder()
        self._frq2 = _eggwrapper.FreqHolder()
        self._site = _eggwrapper.SiteHolder()
        self._genosite = _eggwrapper.Genotypes()
        self._genosite_cache = []
        self._genosite_store = []
        self._struct = None # structure (C++) passed as argument to configure()
        self._struct_auto = True

        self._stats = {
            # code        # description                                                 # flag  # toggler
            'ns_site':    ('Number of analyzed samples per site',                       'sd',   []), 
            'ns_site_o':  ('Number of analyzed outgroup samples per site',              'sd',   []), 
            'Aing':       ('Number of alleles in ingroup',                              'sd',   []),
            'Aotg':       ('Number of alleles in outgroup',                             'sd',   []),
            'Atot':       ('Number of alleles in whole dataset',                        'sd',   []),
            'As':         ('Number of singleton alleles',                               'sd',   []),
            'Asd':        ('Number of singleton alleles (derived)',                     'sd',   []),
            'R':          ('Allelic richness',                                          'sd',   []),
            'thetaIAM':   ('Theta estimator based on He & IAM model',                   'sd',   []),
            'thetaSMM':   ('Theta estimator based on He & SMM model',                   'sd',   []),
            'He':         ('Expected heterozygosity',                                   'sd',   []),
            'Ho':         ('Observed heterozygosity',                                   'sd',   []),
            'Fis':        ('Inbreeding coefficient',                                    'sd',   []),
            'maf':        ('Relative frequency of the minority allele',                 'sd',   []),
            'maf_pop':    ('maf per population (minority allele overall)',              'sd',   []),

            'FstWC':      ('Weir and Cockerham for haploid data',                       'sd',   [self._sd.toggle_fstats_haplo]),
            'FistWC':     ('Weir and Cockerham for diploid data',                       'sd',   [self._sd.toggle_fstats_diplo]),
            'FisctWC':    ('Weir and Cockerham for hierarchical structure',             'sd',   [self._sd.toggle_fstats_hier]),

            'Dj':         ('Jost\'s D',                                                 'sd',   [self._sd.toggle_hstats]),
            'Hst':        ('Hudson\'s Hst',                                             'sd',   [self._sd.toggle_hstats]),
            'Gst':        ('Nei\'s Gst',                                                'sd',   [self._sd.toggle_hstats]),
            'Gste':       ('Hedrick\'s Gst\'',                                          'sd',   [self._sd.toggle_hstats]),

            'numSp':      ('Number of population-specific alleles',                     'as',   []),
            'numSpd':     ('Number of population-specific derived alleles',             'as',   []),
            'numShA':     ('Number of shared alleles',                                  'as',   []),
            'numShP':     ('Number of shared segregating alleles',                      'as',   []),
            'numFxA':     ('Number of fixed alleles',                                   'as',   []),
            'numFxD':     ('Number of fixed differences',                               'as',   []),
            'numSp*':     ('Sites with at least 1 pop-specific allele',                 'as',   []),
            'numSpd*':    ('Sites with at least 1 pop-specific derived allele',         'as',   []),
            'numShA*':    ('Sites with at least 1 shared allele',                       'as',   []),
            'numShP*':    ('Sites with at least 1 shared segregating allele',           'as',   []),
            'numFxA*':    ('Sites with at least 1 fixed allele',                        'as',   []),
            'numFxD*':    ('Sites with at least 1 fixed difference',                    'as',   []),

            'lseff':      ('Number of analysed sites',                                  'd1',   []),
            'nsmax':      ('Maximal number of available samples per site',              'd1',   []),
            'S':          ('Number of segregating sites',                               'd1',   []),
            'Ss':         ('Number of sites with only one singleton allele',            'd1',   []),
            'eta':        ('Minimal number of mutations',                               'd1',   []),
            'Pi':         ('Nucleotide diversity',                                      'd1',   []),

            'sites':      ('Index of polymorphic sites',                                'd1',   [self._d1.toggle_site_lists]),
            'singl':      ('Index of sites with at least one singleton allele',         'd1',   [self._d1.toggle_site_lists]),
            'nall':       ('Number of ingroup alleles at polymorphic sites',            'd1',   [self._d1.toggle_site_lists]),
            'frq':        ('Allelic frequencies at polymorphic sites',                  'd1',   [self._d1.toggle_site_lists]),
            'frqp':       ('Population allelic frequencies at polymorphic sites',       'd1',   [self._d1.toggle_site_lists]),

            'lseffo':     ('Number of analysed orientable sites',                       'd1',   [self._d1.toggle_ori_site]),
            'nsmaxo':     ('Maximal number of available samples per orientable site',   'd1',   [self._d1.toggle_ori_site]),
            'sites_o':    ('Index of orientable polymorphic sites',                     'd1',   [self._d1.toggle_site_lists, self._d1.toggle_ori_site]),
            'singl_o':    ('Index of sites with at least one singleton allele',         'd1',   [self._d1.toggle_site_lists, self._d1.toggle_ori_site]),
            'So':         ('Number of segregating orientable sites',                    'd1',   [self._d1.toggle_ori_site]),
            'Sso':        ('Number of orientable sites with only one singleton allele', 'd1',   [self._d1.toggle_ori_site]),
            'nsingld':    ('Number of derived singletons',                              'd1',   [self._d1.toggle_ori_site]),
            'etao':       ('Minimal number of mutations are orientable sites',          'd1',   [self._d1.toggle_ori_site]),
            'nM':         ('Number of sites available for MFDM test',                   'd1',   [self._d1.toggle_ori_site]),
            'pM':         ('P-value of MDFM test',                                      'd1',   [self._d1.toggle_ori_site]),

            'nseffo':     ('Average number of exploitable samples at orientable sites', 'd1',   [self._d1.toggle_ori_div]),
            'thetaPi':    ('Pi using orientable sites',                                 'd1',   [self._d1.toggle_ori_div]),
            'thetaH':     ('Fay and Wu\'s estimator of theta',                          'd1',   [self._d1.toggle_ori_div]),
            'thetaL':     ('Zeng et al.\'s estimator of theta',                         'd1',   [self._d1.toggle_ori_div]),
            'Hns':        ('Fay and Wu\'s H (unstandardized)',                          'd1',   [self._d1.toggle_ori_div]),
            'Hsd':        ('Fay and Wu\'s H (standardized)',                            'd1',   [self._d1.toggle_ori_div]),
            'E':          ('Zeng et al.\'s E',                                          'd1',   [self._d1.toggle_ori_div]),
            'Dfl':        ('Fu and Li\'s D',                                            'd1',   [self._d1.toggle_ori_div]),
            'F':          ('Fu and Li\'s F',                                            'd1',   [self._d1.toggle_ori_div]),

            'nseff':      ('Average number of exploitable samples',                     'd1',   [self._d1.toggle_basic]),
            'thetaW':     ('Watterson\'s estimator of theta',                           'd1',   [self._d1.toggle_basic]),
            'Dxy':        ('Pairwise distance (if two populations)',                    'd1',   [self._d1.toggle_basic]),
            'Da':         ('Net pairwise distance (if two populations)',                'd1',   [self._d1.toggle_basic]),
            'D':          ('Tajima\'s D',                                               'd1',   [self._d1.toggle_basic]),
            'Deta':       ('Tajima\'s D using eta instead of S',                        'd1',   [self._d1.toggle_basic]),
            'D*':         ('Fu and Li\'s D*',                                           'd1',   [self._d1.toggle_basic]),
            'F*':         ('Fu and Li\'s F*',                                           'd1',   [self._d1.toggle_basic]),

            'R2':         ('Ramos-Onsins and Rozas\'s R2 (using singletons)',           'd2',   [self._d2.toggle_singletons]),
            'R3':         ('Ramos-Onsins and Rozas\'s R3 (using singletons)',           'd2',   [self._d2.toggle_singletons]),
            'R4':         ('Ramos-Onsins and Rozas\'s R4 (using singletons)',           'd2',   [self._d2.toggle_singletons]),
            'Ch':         ('Ramos-Onsins and Rozas\'s Ch (using singletons)',           'd2',   [self._d2.toggle_singletons]),

            'R2E':        ('Ramos-Onsins and Rozas\'s R2E (using external singletons)', 'd2',   [self._d2.toggle_singletons]),
            'R3E':        ('Ramos-Onsins and Rozas\'s R3E (using external singletons)', 'd2',   [self._d2.toggle_singletons]),
            'R4E':        ('Ramos-Onsins and Rozas\'s R4E (using external singletons)', 'd2',   [self._d2.toggle_singletons]),
            'ChE':        ('Ramos-Onsins and Rozas\'s ChE (using external singletons)', 'd2',   [self._d2.toggle_singletons]),

            'B':          ('Wall\'s B statistic',                                       'd2',   [self._d2.toggle_partitions]),
            'Q':          ('Wall\'s Q statistic',                                       'd2',   [self._d2.toggle_partitions]),

            'Kt':         ('Number of haplotypes',                                      'h',    []),
            'Ki':         ('Number of haplotypes (only ingroup)',                       'h',    []),
            'FstH':       ('Hudson\'s Fst',                                             'h',    [self._toggle_haplotype_stats]),
            'Kst':        ('Hudson\'s Kst',                                             'h',    [self._toggle_haplotype_stats]),
            'Snn':        ('Hudson\'s nearest nearest neighbour statistic',             'h',    [self._toggle_haplotype_stats]),

            'rD':         ('R_bar{d} statistic',                                        'rD',   []),

            'Rmin':       ('Minimal number of recombination events',                    'LD',   [self._ld.toggle_Rmin]),
            'RminL':      ('Number of sites used to compute Rmin',                      'LD',   [self._ld.toggle_Rmin]),
            'Rintervals': ('List of start/end positions of recombination intervals',    'LD',   [self._ld.toggle_Rmin]),
            'nPairs':     ('Number of allele pairs used for ZnS, Z*nS, and Z*nS*',      'LD',   [self._ld.toggle_stats]),
            'nPairsAdj':  ('Allele pairs at adjacent sites (used for ZZ and Za)',       'LD',   [self._ld.toggle_stats]),
            'ZnS':        ('Kelly et al.\'s ZnS',                                       'LD',   [self._ld.toggle_stats]),
            'Z*nS':       ('Kelly et al.\'s Z*nS',                                      'LD',   [self._ld.toggle_stats]),
            'Z*nS*':      ('Kelly et al.\'s Z*nS*',                                     'LD',   [self._ld.toggle_stats]),
            'Za':         ('Rozas et al.\'s Za',                                        'LD',   [self._ld.toggle_stats]),
            'ZZ':         ('Rozas et al.\'s ZZ',                                        'LD',   [self._ld.toggle_stats]),

            'Fs':         ('Fu\'s Fs',                                                  'Fs',   []),

            'V':          ('Allele size variance',                                      'V',    []),
            'Ar':         ('Allele range',                                              'V',    []),
            'M':          ('Garza and Williamson (2001)\'s M = Aing / (Ar+1)',          'V',    []),
            'Rst':        ('Slatkin (1995)\'s Rst',                                     'V',    [])
        }

        self.clear_stats()
        self.configure(*args, **kwargs)

    def reset(self):
        """
        Reset all currently computed statistics. Keep the list of
        statistics to compute.
        """
        self._sd.reset()
        self._sd2.reset()
        self._as.reset()
        self._d1.reset_stats()
        self._d2.reset_stats()
        self._h.reset_stats()
        self._rd.reset_stats()
        self._ld.reset()
        self._cv.reset()
        self._site_index = 0
        self._static_sites = True # set to False if sites cannot be considered to be static (required for LD stats)
        self._genosite_store.extend(self._genosite_cache)
        del self._genosite_cache[:]
        if self._struct_auto: self._struct = None

    def configure(self,
                  struct = None,
                  multi = False,
                  multi_hits = False,
                  maf = 0.0,
                  LD_min_n = 2,
                  LD_max_maj = 1.0,
                  LD_multiallelic = 0,
                  LD_min_freq = 0,
                  Rmin_oriented = False):
        """
        Configure the instance. The values provided for parameters will
        affect all subsequent analyses. Calling this method resets all
        statistics.

        :param struct: a :class:`.Structure` instance describing the
            structure of objects that will be passed.

        :param multi: process several alignments, set of sites, or sites
            in a row and only yield statistics when :meth:`.results` is
            called (each of the method considered will return ``None``).

        :param multi_hits: allow multiple mutations at the same site
            (some statistics, like Rmin or Wall's B and Q, never
            consider multiple mutations regardless of this option).

        :param maf: minimum minor allele frequency (sites for which the
            relative minority allele frequency is lower than this will
            be excluded).

        :param LD_min_n: Minimal number of non-missing samples. Allows to
            specify a more stringent filter than *max_missing*.
            Only considered for calculating Rozas *et al.*'s and Kelly's
            statistics (the most stringent of the two criteria applies).

        :param LD_max_maj: Maximal relative frequency of the main allele.
            Only considered for calculating Rozas *et al.*'s and Kelly's
            statistics.

        :param LD_multiallelic: One of 0 (ignore them), 1 (use main
            allele only), and 2 (use all possible pairs of alleles).
            Defines what is done for pairs of sites of which one or
            both have more than two alleles (while computing linkage
            disequilibrium). In case of option 2, a filter can be
            applied with option *LD_min_freq*. Only considered for
            calculating Rozas *et al.*'s and Kelly's statistics.

        :param LD_min_freq: Only considered if option 2 is used for
            *LD_multiallelic*. Only consider alleles that are in
            absolute frequency equal to or larger than the given value.
            Only considered for calculating Rozas *et al.*'s and Kelly's
            statistics

        :param Rmin_oriented: Only for computing Rmin: use only
            orientable sites.
        """

        self.set_structure(struct) # also resets
        if LD_min_n < 0: raise ValueError('LD_min_n is too small')
        self._LD_min_n = LD_min_n
        if LD_max_maj <= 0.0 or LD_max_maj > 1.0: raise ValueError('LD_max_maj out of range')
        self._LD_max_maj = LD_max_maj
        if LD_multiallelic < 0: raise ValueError('LD_multiallelic cannot be negative')
        try: self._LD_multiallelic = [_eggwrapper.MatrixLD.ignore,
                                      _eggwrapper.MatrixLD.use_main,
                                      _eggwrapper.MatrixLD.use_all][LD_multiallelic]
        except IndexError: raise ValueError('LD_multiallelic out of range')
        if LD_min_freq < 0: raise ValueError('LD_min_freq out of range')
        self._LD_min_freq = LD_min_freq
        self._Rmin_oriented = Rmin_oriented
        self._multi_hits = bool(multi_hits)
        self._multi = bool(multi)
        self._d1.set_option_multiple(multi_hits)
        self._d2.set_option_multiple(multi_hits)
        self._sd.set_maf(maf)
        self._sd2.set_maf(maf)
        self._cv.set_maf(maf)

    def set_structure(self, struct):
        """
        Set the structure object. Reset statistics but don't change the
        list of statistics to compute and don't change over parameter
        values.
        """
        self.reset()
        if struct is None:
            self._struct_auto = True
            self._struct = None
            self._aux = None
        else:
            self._struct_auto = False
            self._struct = struct._obj
            self._aux = struct.make_auxiliary()._obj
            self._aux_sorted = struct.make_sorted_auxiliary()._obj
            self._h.setup(struct._obj)
            self._rd.configure(struct._obj)
            self._d2.set_structure(self._aux)
            self._ld.set_structure(self._aux)

    def _check_struct(self, ns):
        if self._struct_auto or self._struct is None:
            struct = _eggwrapper.StructureHolder()
            struct.mk_dummy_structure(ns, 1)
            if self._struct is None: self._h.setup(struct)
            else: self._h.set_structure(struct) # does not reset stats
            self._d2.set_structure(struct)
            self._rd.configure(struct)
            self._ld.set_structure(struct) # no difference between struct and aux
            self._struct = struct
            self._aux = struct # identical because no actual structure
            self._aux_sorted = struct # idem
        else:
            if self._struct.get_req() > ns: raise ValueError('structure does not match number of samples')

    def add_stats(self, *stats):
        """
        add_stats(stat, ...)

        Add one or more statistics to compute. Every statistic identifier must be among
        the list of available statistics, regardless of what data is to
        be analyzed. If statistics cannot be computed, they will be
        returned as ``None``.
        """
        for stat in stats:
            if stat not in self._stats:
                raise ValueError('invalid statistic: {0}'.format(stat))
            self._wanted_stats.add(stat)
            self._flags[self._stats[stat][1]] = True;
            for f in self._stats[stat][2]: f()
        if self._flags['Fs']:
            self._flags['d1'] = True
            self._flags['h'] = True
        if (self._flags['as'] or self._flags['d1'] or self._flags['d2']
                              or self._flags['rD'] or self._flags['LD']
                              or self._flags['h']):
            self._flags['sd'] = True

    def all_stats(self):
        """
        Add all possible statistics. Those who cannot be computed will
        be reported as ``None``. Also reset all currently computed statistics (if any).
        """
        self.add_stats(*self._stats)

    def clear_stats(self):
        """
        Clear the list of statistics to compute.
        """
        self._wanted_stats = set()
        self._flags = dict.fromkeys([self._stats[k][1] for k in self._stats], False)
        self._sd.toggle_off()
        self._sd2.toggle_off()
        self._d1.toggle_off()
        self._ld.toggle_off()
        self._toggle_off()

    def _toggle_off(self):
        self._flag_haplotype_stats = False

    def _toggle_haplotype_stats(self):
        self._flag_haplotype_stats = True

    def process_freq(self, frq, position=None):
        """
        Analyze already computed frequencies.

        :param frq: a :class:`.Freq` instance.
        :param position: position of the site. Must be an integer value.
            By default, use the site loading index.
        :return: A dictionary of statistics, unless *multi* is set.
        """

        if not self._multi:
            stats = dict.fromkeys(self._wanted_stats, None)

        if position is None: position = self._site_index

        if self._flags['sd'] == True:
            res = self._sd.process(frq._obj)
            if not self._multi: self._get_sd_stats(res, stats)

        if self._flags['as'] == True and (res & 1024) != 0 and self._sd.npop_eff1() > 1:
            self._as.process(frq._obj)
            if not self._multi: self._get_as_stats(stats)

        if self._flags['d1'] == True and (res & 2) != 0:
            self._d1.load(frq._obj, self._sd, position)

        if self._flags['V'] == True:
            if frq._alphabet.type not in ['int', 'range']: raise ValueError('cannot compute V, Ar, M, and Rst with this alphabet')
            b = self._cv.compute(frq._obj, frq._alphabet._obj)
            if not self._multi and b:
                if 'V' in stats: stats['V'] = self._cv.curr_V()
                if 'Ar' in stats: stats['Ar'] = self._cv.curr_Ar()
                if 'M' in stats and self._cv.curr_M() != _eggwrapper.UNDEF: stats['M'] = self._cv.curr_M()
                if 'Rst' in stats and self._cv.curr_Rst() != _eggwrapper.UNDEF: stats['Rst'] = self._cv.curr_Rst()

        self._site_index += 1
        if not self._multi: return stats

    def process_site(self, site, position=None):
        """
        Analyze a site.

        :param site: a :class:`.Site` instance.
        :param position: position of the site. Must be an integer value.
            By default, use the site index.
        :return: A dictionary of statistics, unless *multi* is set.
        """

        self._check_struct(site._obj.get_ns())
        if position is None: position = self._site_index
        if not self._multi: stats = dict.fromkeys(self._wanted_stats, None)
        if self._flags['d2'] or self._flags['h']: # add LD for process_sites/process_align
            self._genosite.process(site._obj, self._struct)
            if self._flags['d2']:
                self._frq2.setup_structure(self._aux)
                self._frq2.process_site(self._genosite.site())
                self._sd2.process(self._frq2)
        self._frq.setup_structure(self._struct)
        self._frq.process_site(site._obj)
        if self._flags['sd'] == True:
            res = self._sd.process(self._frq)
            if not self._multi: self._get_sd_stats(res, stats)
        if self._flags['as'] == True and (res & 1024) != 0 and self._sd.npop_eff1() > 1:
            self._as.process(self._frq)
            if not self._multi: self._get_as_stats(stats)
        if self._flags['d1'] == True and (res & 2) != 0:
            self._d1.load(self._frq, self._sd, position)
        if self._flags['d2'] == True and (res & 2) != 0:
            self._d2.load(self._genosite.site(), self._sd2, self._frq2)
        if self._flags['h'] == True and self._sd.Aglob() > 1 and (self._sd.Aing() == 2 or self._multi_hits == True):
            self._h.load(self._genosite)
        if self._flags['rD'] == True:
            self._rd.load(site._obj)
        if self._flags['V'] == True:
            if site._alphabet.type not in ['int', 'range']: raise ValueError('cannot compute V, Ar, M, and Rst with this alphabet')
            b = self._cv.compute(self._frq, site._alphabet._obj)
            if not self._multi and b:
                if 'V' in stats: stats['V'] = self._cv.curr_V()
                if 'Ar' in stats: stats['Ar'] = self._cv.curr_Ar()
                if 'M' in stats and self._cv.curr_M() != _eggwrapper.UNDEF: stats['M'] = self._cv.curr_M()
                if 'Rst' in stats and self._cv.curr_Rst() != _eggwrapper.UNDEF: stats['Rst'] = self._cv.curr_Rst()
        self._site_index += 1
        if self._multi: self._static_sites = False
        else: return stats

    def process_sites(self, sites, positions=None):
        """
        Analyze a list of sites.

        :param sites: a list, or other sequence of :class:`.Site`
            instance. (:class:`.VcfWindow` instances
            with a number of sites greater than 0 are also supported.)
        :param positions: a list, or other sequence of positions for all
            sites, or ``None``. If ``None``, use the index of each site. Otherwise,
            must be a sequences of integer values (length matching the
            length of *sites*).
        :return: A dictionary of statistics, unless *multi* is set.
        """

        if positions is None:
            positions = (self._site_index + i for i in range(len(sites)))
        elif len(positions) != len(sites):
            raise ValueError('number of positions must match the number of sites')
        if len(sites) > 0:
            for site, pos in zip(sites, positions):
                self._check_struct(site._obj.get_ns())
                self._frq.setup_structure(self._struct)
                self._frq.process_site(site._obj)
                keep = False
                if self._flags['d2'] or self._flags['h'] or self._flags['LD']:
                    self._genosite.process(site._obj, self._struct)
                    if self._flags['d2']:
                        self._frq2.setup_structure(self._aux)
                        self._frq2.process_site(self._genosite.site())
                        self._sd2.process(self._frq2)
                if self._flags['sd'] == True: res = self._sd.process(self._frq)  # don't require consistency
                if self._flags['as'] == True and (res & 1024) != 0 and self._sd.npop_eff1() > 1: self._as.process(self._frq)  # don't require consistency
                if self._flags['d1'] == True and (res & 2) != 0: self._d1.load(self._frq, self._sd, pos)  # don't require consistency
                if self._flags['d2'] == True and (res & 2) != 0:
                    self._d2.load(self._genosite.site(), self._sd2, self._frq2)
                if self._flags['h'] == True and self._sd.Aglob() > 1 and (self._multi_hits == True or self._sd.Aing() < 3):
                    self._h.load(self._genosite)
                    keep |= True
                if self._flags['rD'] == True and (res & 2) != 0 and self._sd.Aing() > 1: self._rd.load(site._obj)
                if self._flags['LD'] == True and (res & 2) != 0 and (self._sd.Aing() == 2 or (self._sd.Aing() > 2 and self._LD_multiallelic != _eggwrapper.MatrixLD.ignore)):
                    keep |= True
                    self._ld.load(self._genosite, pos)
                if self._flags['V'] == True:
                    if site._alphabet.type not in ['int', 'range']: raise ValueError('cannot compute V, Ar, M, and Rst with this alphabet')
                    self._cv.compute(self._frq, site._alphabet._obj)

                # if need to keep site, put it in cache and claim new
                if keep:
                    self._genosite_cache.append(self._genosite)
                    self._genosite = self._get_genosite_from_store()

        self._site_index += len(sites)
        if self._multi: self._static_sites = False # prevent attempting using sites (which may be invalidated)
        else: return self.results()

    def process_align(self, align, positions=None, max_missing=0.0):
        """
        Analyze an alignment.

        :param align: an :class:`.Align` instance.
        :param positions: a list, or other sequence of positions for all
            sites, or ``None``. If ``None``, use the index of each site. Otherwise,
            must be a sequences of integer values (length matching the
            number of sites).
        :param max_missing: Maximum relative proportion of missing data (excluding
            outgroup). The default is to exclude all sites with missing data. Missing
            data include all ambiguity characters and alignment gaps.
            Sites not passing this threshold are ignored for all
            analyses.

            .. note::
                Here, *max_missing* is a relative proportion to allow
                using the same value for different alignments that might
                not have the same number of samples (to avoid reevaluating
                *max_missing* if the user wants the same maximum rate
                of missing data). In other functions, *max_missing* is
                the maximum *number* of missing and is an integer.

        :return: A dictionary of statistics, unless *multi* is set.
        """
        if not isinstance(align, _interface.Align): raise TypeError('expect an Align instance')

        # check max_missing
        max_missing = float(max_missing)
        if max_missing < 0.0 or max_missing > 1.0:
            raise ValueError('invalid value for `max_missing` argument: {0}'.format(max_missing))

        # get number of sites and positions
        ls = align._obj.get_nsit_all()
        if positions is None:
            positions = (self._site_index + i for i in range(ls))
        elif len(positions) != ls:
            raise ValueError('number of positions must match the number of sites')

        self._check_struct(align._obj.get_nsam())
        ns = self._struct.get_ni()
        no = self._struct.get_no()

        # get max_missing number
        min_valid = math.floor(ns - max_missing * ns)

        # process sites
        for idx, pos in enumerate(positions):
            self._site.reset()
            if self._site.process_align(align._obj, idx, self._struct) < min_valid:
                continue

            if self._flags['d2'] or self._flags['h'] or self._flags['LD']:
                self._genosite.process(self._site, self._aux_sorted)
                if self._flags['d2']:
                    self._frq2.setup_structure(self._aux)
                    self._frq2.process_site(self._genosite.site())
                    self._sd2.process(self._frq2)

            self._frq.setup_structure(self._aux_sorted)
            self._frq.process_site(self._site)

            keep = False
            if self._flags['sd'] == True: res = self._sd.process(self._frq)
            if self._flags['as'] == True and (res & 1024) != 0 and self._sd.npop_eff1() > 1: self._as.process(self._frq)

            if self._flags['d1'] == True and (res & 2) != 0: self._d1.load(self._frq, self._sd, pos)
            if self._flags['d2'] == True and (res & 2) != 0:
                self._d2.load(self._genosite.site(), self._sd2, self._frq2)
            if self._flags['h'] == True and self._sd.Aglob() > 1 and (self._multi_hits == True or self._sd.Aing() < 3):
                self._h.load(self._genosite)
                keep |= True
            if self._flags['rD'] == True and (res & 2) != 0 and self._sd.Aing() > 1:
                self._rd.load(self._site) # Rd does not assume that data are phased
            if self._flags['LD'] == True and (res & 2) != 0 and (self._sd.Aing() == 2 or (self._sd.Aing() > 2 and self._LD_multiallelic != _eggwrapper.MatrixLD.ignore)):
                self._ld.load(self._genosite, pos)
                keep |= True
            if self._flags['V'] == True:
                if align._alphabet.type not in ['int', 'range']: raise ValueError('cannot compute V, Ar, M, and Rst with this alphabet')
                self._cv.compute(self._frq, align._alphabet._obj)

            # if need to keep site, put it in cache and claim new
            if keep:
                self._genosite_cache.append(self._genosite)
                self._genosite = self._get_genosite_from_store()

        self._site_index += ls
        if not self._multi: return self.results()

    def results(self):
        """
        Get computed statistics.
        Return the value of statistics for all sites since the last call
        to this method, to :meth:`reset`, or any addition of statistics,
        or the object creation,
        whichever is most recent. For statistics that can not be
        computed, ``None`` is returned.
        """

        stats = dict.fromkeys(self._wanted_stats, None)
        if self._flags['sd'] == True:
            res = self._sd.average()
            self._get_sd_stats(res, stats)
        if self._flags['as'] == True:
            self._as.total()
            if (res & 1024) != 0 and self._sd.npop_eff1() > 1:
                self._get_as_stats(stats)
                if self._as.nsites() > 0:
                    if 'numSp*' in stats: stats['numSp*'] = self._as.Sp_T1()
                    if 'numSpd*' in stats and self._as.nsites_o() > 0: stats['numSpd*'] = self._as.Spd_T1()
                    if 'numShP*' in stats: stats['numShP*'] = self._as.ShP_T1()
                    if 'numShA*' in stats: stats['numShA*'] = self._as.ShA_T1()
                    if 'numFxD*' in stats: stats['numFxD*'] = self._as.FxD_T1()
                    if 'numFxA*' in stats: stats['numFxA*'] = self._as.FxA_T1()

        if self._flags['d1'] == True:
            res = self._d1.compute()
            if 'lseff'  in stats: stats['lseff'] =  self._d1.ls()
            if 'lseffo'  in stats: stats['lseffo'] =   self._d1.lso()
            if (res & 1) != 0:
                if 'nsmax'   in stats: stats['nsmax'] =    self._d1.nsmax()
                if 'S'       in stats: stats['S'] =        self._d1.S()
                if 'Ss'      in stats: stats['Ss'] =       self._d1.Ss()
                if 'sites'   in stats: stats['sites'] =    [self._d1.site(i) for i in range(self._d1.S())]
                if 'singl'   in stats: stats['singl'] =    [self._d1.singl(i) for i in range(self._d1.Ss())]
                if 'nall'    in stats: stats['nall'] =     [self._d1.nall(i) for i in range(self._d1.S())]
                if 'frq'     in stats: stats['frq'] =      [[self._d1.frq(i, j) for j in range(self._d1.nall(i))] for i in range(self._d1.S())]
                if 'frqp'    in stats: stats['frqp'] =     [[[self._d1.frqp(i, j, k) for k in range(self._struct.num_pop())] for j in range(self._d1.nall(i))] for i in range(self._d1.S())]
                if 'eta'     in stats: stats['eta'] =      self._d1.eta()
                if 'Pi'      in stats: stats['Pi'] =       self._d1.Pi()
            if (res & 8) != 0:
                if 'nsmaxo'  in stats: stats['nsmaxo'] =   self._d1.nsmaxo()
                if 'So'      in stats: stats['So'] =       self._d1.So()
                if 'Sso'     in stats: stats['Sso'] =      self._d1.Sso()
                if 'nsingld' in stats: stats['nsingld'] =  self._d1.nsingld()
                if 'sites_o' in stats: stats['sites_o'] =  [self._d1.site_o(i) for i in range(self._d1.So())]
                if 'singl_o' in stats: stats['singl_o'] =  [self._d1.singl_o(i) for i in range(self._d1.Sso())]
                if 'etao'    in stats: stats['etao'] =     self._d1.etao()
                if 'nM'      in stats: stats['nM'] =       self._d1.nM()
            if (res & 16) != 0:
                if 'pM'      in stats: stats['pM'] =       self._d1.pM()
            if (res & 32) != 0:
                if 'nseffo'  in stats: stats['nseffo'] =   self._d1.nseffo()
                if 'thetaPi' in stats: stats['thetaPi'] =  self._d1.thetaPi()
                if 'thetaH'  in stats: stats['thetaH'] =   self._d1.thetaH()
                if 'thetaL'  in stats: stats['thetaL'] =   self._d1.thetaL()
                if 'Hns'     in stats: stats['Hns'] =      self._d1.Hns()
            if (res & 1024) != 0:
                if 'Hsd'     in stats: stats['Hsd'] =      self._d1.Hsd()
            if (res & 2048) != 0:
                if 'E'       in stats: stats['E'] =        self._d1.E()
            if (res & 4096) != 0:
                if 'Dfl'     in stats: stats['Dfl'] =      self._d1.Dfl()
            if (res & 8192) != 0:
                if 'F'       in stats: stats['F'] =        self._d1.F()
            if (res & 128) != 0:
                if 'nseff'   in stats: stats['nseff'] =    self._d1.nseff()
                if 'thetaW'  in stats: stats['thetaW'] =   self._d1.thetaW()
            if (res & 16384) != 0:
                if 'Dxy'     in stats: stats['Dxy'] =      self._d1.Dxy()
                if 'Da'      in stats: stats['Da'] =       self._d1.Da()
            if (res & 256) != 0:
                if 'F*'      in stats: stats['F*'] =       self._d1.Fstar()
            if (res & 512) != 0:
                if 'D'       in stats: stats['D'] =        self._d1.D()
                if 'Deta'    in stats: stats['Deta'] =     self._d1.Deta()
                if 'D*'      in stats: stats['D*'] =       self._d1.Dstar()

        if self._flags['d2'] == True:
            res = self._d2.compute()
            if (res & 1) == 0:
                if (res & 256) != 0:
                    if 'R2' in stats: stats['R2'] = self._d2.R2()
                    if 'R3' in stats: stats['R3'] = self._d2.R3()
                    if 'R4' in stats: stats['R4'] = self._d2.R4()
                    if 'Ch' in stats: stats['Ch'] = self._d2.Ch()
                if (res & 512) != 0:
                    if 'R2E' in stats: stats['R2E'] = self._d2.R2E()
                    if 'R3E' in stats: stats['R3E'] = self._d2.R3E()
                    if 'R4E' in stats: stats['R4E'] = self._d2.R4E()
                    if 'ChE' in stats: stats['ChE'] = self._d2.ChE()
                if (res & 1024) != 0:
                    if 'B' in stats: stats['B'] = self._d2.B()
                    if 'Q' in stats: stats['Q'] = self._d2.Q()

        if self._flags['h'] == True and self._h.n_sites() > 0 and not self._h.invalid():
            self._h.cp_haplotypes()
            if 'Kt' in stats and (self._h.ne_ing() > 0 or self._h.ne_otg() > 0): stats['Kt'] = self._h.ng_hapl()
            if self._h.ne_ing() > 0 and 'Ki' in stats: stats['Ki'] = self._h.ni_hapl()
            if self._flag_haplotype_stats:
                self._h.cp_dist()
                res = self._h.cp_stats()
                if 'FstH' in stats and res & 1 != 0: stats['FstH'] = self._h.Fst()
                if 'Kst' in stats and res & 2 != 0: stats['Kst'] = self._h.Kst()
                if 'Snn' in stats:
                    stats['Snn'] = self._h.Snn()
                    if stats['Snn'] == _eggwrapper.UNDEF: stats['Snn'] = None

        if 'Fs' in stats and self._h.n_sites() and self._h.ni_hapl() > 0:
            stats['Fs'] = _eggwrapper.Fs(self._h.ne_ing(), self._h.ni_hapl(), self._d1.Pi())

        if self._static_sites == True and 'rD' in stats:
            stats['rD'] = self._rd.compute()
            if stats['rD'] == _eggwrapper.UNDEF: stats['rD'] = None

        if self._static_sites == True and self._flags['LD'] == True:
            flag = self._ld.process(self._LD_min_n, self._LD_max_maj,
                                    self._LD_multiallelic, self._LD_min_freq,
                                    self._Rmin_oriented)
            if 'nPairs' in stats: stats['nPairs'] = self._ld.num_allele_pairs()
            if 'nPairsAdj' in stats: stats['nPairsAdj'] = self._ld.num_allele_pairs_adj()
            if 'RminL' in stats: stats['RminL'] = self._ld.Rmin_num_sites()
            if (flag&1) != 0:
                if 'ZnS' in stats: stats['ZnS'] = self._ld.ZnS()
                if 'Z*nS' in stats: stats['Z*nS'] = self._ld.ZnS_star1()
                if 'Z*nS*' in stats: stats['Z*nS*'] = self._ld.ZnS_star2()
            if (flag&2) != 0:
                if 'Za' in stats: stats['Za'] = self._ld.Za()
                if 'ZZ' in stats: stats['ZZ'] = self._ld.ZZ()
            if (flag&4) != 0:
                if 'Rmin' in stats: stats['Rmin'] = self._ld.Rmin()
                if 'Rintervals' in stats: stats['Rintervals'] = [(self._ld.Rmin_left(i), self._ld.Rmin_right(i)) for i in range(self._ld.Rmin())]

        if self._flags['V']:
            if self._cv.num_sites() > 0:
                if 'V' in stats: stats['V'] = self._cv.average_V()
                if 'Ar' in stats: stats['Ar'] = self._cv.average_Ar()
                if 'M' in stats and self._cv.num_sites_m() > 0: stats['M'] = self._cv.average_M()
                if 'Rst' in stats and self._cv.average_Rst() != _eggwrapper.UNDEF: stats['Rst'] = self._cv.average_Rst()

        self.reset()
        return stats

    def _get_sd_stats(self, flag, stats):
        if (flag & 1) != 0 and 'ns_site' in stats:
            if (flag & 512) != 0: stats['ns_site'] = int(self._sd.ns())
            else: stats['ns_site'] = self._sd.ns()
        if (flag & 1) != 0 and 'ns_site_o' in stats:
            if (flag & 512) != 0: stats['ns_site_o'] = int(self._sd.nso())
            else: stats['ns_site_o'] = self._sd.nso()
        if (flag & 2) != 0:
            if 'Aing' in stats:
                if (flag & 512) != 0: stats['Aing'] = int(self._sd.Aing())
                else: stats['Aing'] = self._sd.Aing()
            if 'Atot' in stats:
                if (flag & 512) != 0: stats['Atot'] = int(self._sd.Aglob())
                else: stats['Atot'] = self._sd.Aglob()
            if 'As' in stats:
                if (flag & 512) != 0: stats['As'] = int(self._sd.S())
                else: stats['As'] = self._sd.S()
            if 'R' in stats: stats['R'] = self._sd.R()
            if 'He' in stats: stats['He'] = self._sd.He()

        if (flag & 4) != 0:
            if 'thetaIAM' in stats: stats['thetaIAM'] = self._sd.thetaIAM()
            if 'thetaSMM' in stats: stats['thetaSMM'] = self._sd.thetaSMM()

        if (flag & 8) != 0:
            if 'Ho' in stats: stats['Ho'] = self._sd.Ho()
            if 'Fis' in stats and self._sd.He() > 0.0:
                stats['Fis'] = 1 - self._sd.Ho() / self._sd.He()

        if (flag & 16) != 0:
            if 'Asd' in stats:
                if (flag & 512) != 0: stats['Asd'] = int(self._sd.Sd())
                else: stats['Asd'] = self._sd.Sd()
        if (flag & 16384) != 0:
            if 'Aotg' in stats:
                if (flag & 512) != 0: stats['Aotg'] = int(self._sd.Aout())
                else: stats['Aotg'] = self._sd.Aout()

        if (flag & 32) != 0 and 'FstWC' in stats:
            n = self._sd.n()
            d = self._sd.d()
            stats['FstWC'] = n/d if d > 0.0 else None

        if (flag & 64) != 0 and 'FistWC' in stats and self._struct.get_ploidy() > 1:
            a = self._sd.a()
            b = self._sd.b()
            c = self._sd.c()
            stats['FistWC'] = ( 1.0 - c/(b+c) if (b+c) > 0.0 else None,
                a/(a+b+c), 1.0 - c/(a+b+c)) if (a+b+c) > 0.0 else None

        if (flag & 128) != 0 and 'FisctWC' in stats and self._struct.get_ploidy() > 1:
            a0 = self._sd.a0()
            b2 = self._sd.b2()
            b1 = self._sd.b1()
            c0 = self._sd.c0()
            stats['FisctWC'] = ( 1.0 - c0/(b1+c0) if (b1+c0) > 0.0 else None,
                    (a0+b2) / (a0+b2+b1+c0), a0/(a0+b2+b1+c0),
                    1.0 - c0/(a0+b2+b1+c0)) if (a0+b2+b1+c0) > 0.0 else None

        if (flag & 256) != 0:
            if 'Dj' in stats: stats['Dj'] = self._sd.D()
            if 'Hst' in stats and (flag & 2048) != 0: stats['Hst'] = self._sd.Hst()
            if 'Gst' in stats and (flag & 4096) != 0: stats['Gst'] = self._sd.Gst()
            if 'Gste' in stats and (flag & 8192) != 0: stats['Gste'] = self._sd.Gste()

        if (flag & 1024) != 0:
            if 'maf' in stats: stats['maf'] = self._sd.maf()
            if 'maf_pop' in stats: stats['maf_pop'] = [self._sd.maf_pop(i) for i in range(self._sd.num_pop())]

    def _get_as_stats(self, stats):
        if self._as.nsites() > 0:
            if 'numSp' in stats: stats['numSp'] = self._as.Sp()
            if 'numSpd' in stats and self._as.nsites_o() > 0: stats['numSpd'] = self._as.Spd()
            if 'numShP' in stats: stats['numShP'] = self._as.ShP()
            if 'numShA' in stats: stats['numShA'] = self._as.ShA()
            if 'numFxD' in stats: stats['numFxD'] = self._as.FxD()
            if 'numFxA' in stats: stats['numFxA'] = self._as.FxA()



"""
    Copyright 2008-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from . import tools
from . import random
from . import stats
from . import io
from . import coalesce
from . import wrappers
from ._interface import Align, Container, SampleView, SequenceView, LabelView, encode, Structure, struct_from_labels, struct_from_dict, struct_from_samplesizes, struct_from_iterable
from ._site import Site, site_from_align, site_from_list, site_from_vcf
from ._tree import Tree, Node
from .alphabets import Alphabet
from ._freq import Freq, freq_from_site, freq_from_list, freq_from_vcf

__version__ = '3.2.0'



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import os, egglib, sys, unittest, random, re, gc, time
from operator import truediv
import collections

mk_dict1=({None: {'0': {'0': [0, 1], '1': [2, 3], '2': [4, 5], '3': [6, 7]}}}, {})
mk_dict2=({None: {'0': {'0': [0], '1': [1], '2': [2], '3': [3], '4': [4], '5': [5], '6': [6], '7': [7]}}}, {})
mk_dict3=({None: {'0': {'0': [0], '1': [1], '2': [2], '3': [3], '4': [4], '5': [5], '6': [6], '7': [7], '8': [8], '9': [9], '10': [10], '11': [11], '12': [12], '13': [13], '14': [14], '15': [15], '16': [16], '17': [17], '18': [18], '19': [19]},
                        '1': {'32': [32], '33': [33], '34': [34], '35': [35], '36': [36], '37': [37], '38': [38], '39': [39], '20': [20], '21': [21], '22': [22], '23': [23], '24': [24], '25': [25], '26': [26], '27': [27], '28': [28], '29': [29], '30': [30], '31': [31]},
                        '2': {'40': [40], '41': [41], '42': [42], '43': [43], '44': [44], '45': [45], '46': [46], '47': [47], '48': [48], '49': [49], '50': [50], '51': [51], '52': [52], '53': [53], '54': [54], '55': [55], '56': [56], '57': [57], '58': [58], '59': [59]},
                        '3': {'64': [64], '65': [65], '66': [66], '67': [67], '68': [68], '69': [69], '70': [70], '71': [71], '72': [72], '73': [73], '74': [74], '75': [75], '76': [76], '77': [77], '78': [78], '79': [79], '60': [60], '61': [61], '62': [62], '63': [63]},
                        '4': {'80': [80], '91': [91], '92': [92]},
                        '5': {'81': [81], '82': [82], '83': [83], '84': [84], '85': [85], '86': [86], '87': [87], '88': [88], '89': [89], '90': [90]}}}, {})

class ParamDict_test(unittest.TestCase):

    def setUp(self):
        self.coal= egglib.coalesce.Simulator(num_pop=4)
        self.params = self.coal.params

    def tearDown(self):
            del self.coal

    def test_ParamDict_T(self):
        self.assertIsInstance(self.params, egglib.coalesce._param_helpers.ParamDict)

    def test_ParamDict_E(self):
        with self.assertRaises(TypeError):
            egglib.coalesce.Simulator(None)

    def test_disable_trans_matrix_T(self):
        self.params.update(num_sites=10, theta=5.0, recomb=5.0, num_mut=0,mut_model='IAM', num_alleles=4, rand_start=True)
        self.params['trans_matrix'][0, 1] = 1.2
        self.params['trans_matrix'][1, 0] = 1.2
        self.params['trans_matrix'][2, 3] = 1.2
        self.params['trans_matrix'][3, 2] = 1.2
        self.params.disable_trans_matrix()
        self.assertEqual(self.params['trans_matrix'][0,2],1.0)
        self.assertEqual(self.params['trans_matrix'][1,1], None)

    def test_get_values_T(self):
        coal2 =egglib.coalesce.Simulator(num_pop=4)
        params2=coal2.params
        params2.update(num_sites=4)
        params2.update(num_chrom=[1,2,4,5],site_pos=[0.1,0.2,0.3,0.4])
        self.params.get_values(params2)
        self.assertEqual(self.params.summary(), params2.summary())

    def test_get_values_E(self):
        coal =egglib.coalesce.Simulator(num_pop=4)
        cnt= egglib.Container(egglib.alphabets.DNA)
        params=coal.params
        with self.assertRaises(TypeError):
            params.get_values(cnt)

    def test_summary_T(self):
        summ= self.params.summary()
        self.assertIn('Number of populations:', summ)

    def test_set_migr_T(self):
        self.params.set_migr(10)
        self.assertEqual(self.params['migr_matrix'][0,2], truediv(10,(4-1)))

    def test_set_migr_E(self):
        with self.assertRaises(ValueError):
            self.params.set_migr(-10)

    def test_mk_structure_T(self):
        coal = egglib.coalesce.Simulator(1, num_indiv=[2])
        coal.params['events'].add(T=0.5, cat='sample', num_chrom=0, num_indiv=2, idx=0, label='0')
        self.assertIsInstance(coal.params.mk_structure(), egglib.Structure)
        self.assertEqual(coal.params.mk_structure().as_dict(), mk_dict1)
        self.assertEqual(coal.params.mk_structure(skip_indiv=True).as_dict(), mk_dict2)

        coal2 = egglib.coalesce.Simulator(6, num_chrom=[20, 20, 20, 20, 1, 10],
                                                 num_sites=1000, num_alleles=4,
                                             theta=1.5, migr=0.0,
                                             rand_start=True)

        coal2.params['migr_matrix'] = [ [None, 0.75, 0.75, 0.75, 0.00, 0.75],
                            [0.75, None, 0.75, 0.75, 0.00, 0.75],
                            [0.75, 0.75, None, 0.75, 0.00, 0.75],
                            [0.75, 0.75, 0.75, None, 0.00, 0.75],
                            [0.00, 0.00, 0.00, 0.00, None, 0.00],
                            [0.75, 0.75, 0.75, 0.75, 0.00, None] ]
        coal2.params['N'][4] = 4.0
        coal2.params['events'].add(T=4.0, cat='merge', src=5, dst=0)
        coal2.params['events'].add(T=4.0, cat='merge', src=4, dst=0)
        coal2.params['events'].add(T=4.0, cat='merge', src=3, dst=0)
        coal2.params['events'].add(T=4.0, cat='merge', src=2, dst=0)
        coal2.params['events'].add(T=4.0, cat='merge', src=1, dst=0)
        coal2.params['events'].add(T=1.0, cat='sample', idx=4, label='4', num_chrom=2, num_indiv=0)
        self.assertEqual(coal2.params.mk_structure().as_dict(), mk_dict3)

    def test_mk_structure_E(self):
        coal = egglib.coalesce.Simulator(1)
        coal.params.mk_structure()
        coal.params['num_chrom'] = [10]
        coal.params.mk_structure()
        coal.params['num_indiv'] = [2]
        with self.assertRaises(ValueError):
            coal.params.mk_structure()

    def test_keys_T(self):
        list_keys_check=['num_pop', 'num_sites', 'recomb', 'theta', 'num_mut', 'mut_model', 'TPM_proba', 'TPM_param', 'num_alleles', 'rand_start', 'num_chrom', 'num_indiv', 'N', 'G', 's', 'site_pos', 'site_weight', 'migr_matrix', 'trans_matrix', 'events', 'max_iter']
        self.assertEqual(list(self.params.keys()), list_keys_check)

    def test_values_T(self):
        self.assertEqual(len(self.params.values()),21 )

    def test_items_T(self):
        match=('TPM_proba', 0.5)
        self.assertIn(match, self.params.items())
        self.assertEqual(len(self.params), 21)

    def test_has_key_T(self):
        self.assertTrue(self.params.has_key('num_chrom'))
        self.assertFalse(self.params.has_key('FAIL'))

    def test_get_T(self):
        self.assertEqual(self.params.get('TPM_proba'),0.5)
        self.assertIsNone(self.params.get('FAIL'))

    def test_iterkeys_T(self):
        self.assertIsInstance(self.params.keys(), collections.abc.Iterable)

    def test_values_T(self):
        self.assertIsInstance(self.params.values(), collections.abc.Iterable)

    def test__iter__T(self):
        self.assertIsInstance(self.params, collections.abc.Iterable)

    def test_copy_T(self):
        copy=self.params.copy()
        self.assertIsInstance(copy, egglib.coalesce._param_helpers.ParamDict)
        for key in self.params.keys():
            self.assertEqual(str(self.params.get(key)), str(copy.get(key)))

    def test_update_T(self):
        n_mut=self.params.get('num_mut')
        self.params.update({'theta': 7.5, 'mut_model': 'IAM'}, theta=0.0, num_mut=15)
        self.assertNotEqual(self.params.get('num_mut'), n_mut)
        
    def test_update_E(self):
        coal=egglib.coalesce.Simulator(num_pop=4)
        params=coal.params
        with self.assertRaises(KeyError):
            params.update(egglib=4)
        with self.assertRaises(ValueError):
            params.update(other='FAIL', num_sites=4)

    def test__len__T(self):
        self.assertEqual(len(self.params), 21)

    def test__reversed__T(self):
        self.assertIsInstance(reversed(self.params), collections.abc.Iterable)

    def test__contains__T(self):
        self.assertTrue('site_pos' in self.params)

    def test__str__T(self):
        self.assertIsInstance(str(self.params), str)

    def test__getitem__T(self):
        self.assertEqual(self.params['num_pop'],4)

    def test__getitem__E(self):
        with self.assertRaises(KeyError):
            self.params['ERROR']

    def test__setitem__T(self):
        n_all1=self.params['num_alleles']
        self.params['num_alleles']=4
        self.assertNotEqual(n_all1, self.params['num_alleles'])

    def test__setitem__E(self):
        with self.assertRaises(KeyError):
            self.params['ERROR']=100
        with self.assertRaises(ValueError):
            self.params['num_pop']=10
        with self.assertRaises(ValueError):
            self.params['num_sites']=-10
        with self.assertRaises(ValueError):
            self.params['recomb']=-10
        with self.assertRaises(ValueError):
            self.params['theta']=-10
        with self.assertRaises(ValueError):
            self.params['num_mut']=-15
        with self.assertRaises(ValueError):
            self.params['num_mut']=10
            self.params['theta']=10

        with self.assertRaises(ValueError):
            self.params['mut_model']='ERR'
        with self.assertRaises(ValueError):
            self.params['TPM_proba']=150
        with self.assertRaises(ValueError):
            self.params['TPM_param']=-10
        with self.assertRaises(ValueError):
            self.params['num_alleles']=1
        with self.assertRaises(ValueError):
            self.params['max_iter']=-150
        with self.assertRaises(ValueError):
            self.params['events']='error'

    def test_add_event_T(self):
        ne1=len(self.params._events)
        self.params.add_event('sample', T=1.0, idx=0, label='0', num_chrom=5, num_indiv=0)
        ne2=len(self.params._events)
        self.assertTrue(ne1<ne2)

    def test_add_event_E(self):
        with self.assertRaises(ValueError):
            self.params.add_event('error', T=1.0, idx=0, label=0, num_chrom=5, num_indiv=0)
        with self.assertRaises(ValueError):
            self.params.add_event('sample', T=1.0, label=0, num_chrom=5, num_indiv=0)
        with self.assertRaises(ValueError):
            self.params.add_event('merge', T=1.0, src=-150, dst=3)
            self.params.add_event('merge', T=1.0, src=2, dst=10)
            self.params.add_event('merge', T=1.0, src=2, dst=2)

class ParamList_test(unittest.TestCase):
        
    def setUp(self):
        self.coal= egglib.coalesce.Simulator(num_pop=4,num_sites=3, theta=5.0, recomb=5.0, num_mut=0,mut_model='IAM', num_alleles=4, rand_start=True, site_pos=[0.3, 0.5, 0.7])
        self.params = self.coal.params

    def tearDown(self):
            del self.coal

    def test_ParamList_T(self):
        num_chrom=self.params['site_pos']
        self.assertIsInstance(num_chrom, egglib.coalesce._param_helpers.ParamList)

    def test__getitem__T(self):
        self.params['site_pos'][0]
        self.assertEqual(self.params['site_pos'][0],0.3)

    def test__setitem__T(self):
        self.params['site_pos'][0]=0.4
        self.assertEqual(self.params['site_pos'][0],0.4)

    def test__iter__T(self):
        self.assertIsInstance(self.params['site_pos'], collections.abc.Iterable)

    def test__add__T(self):
        sum_sites=self.params['site_pos'][0]+self.params['site_pos'][1]
        self.assertEqual(sum_sites, 0.8)

    def test__radd__T(self):
        sum_sites=self.params['site_pos'][0]+0.6
        self.assertEqual(round(sum_sites,1), 0.9)

    def test__mul__T(self):
        sum_sites=self.params['site_pos'][0]*self.params['site_pos'][1]
        self.assertEqual(sum_sites, 0.15)

    def test__rmul__T(self):
        sum_sites=self.params['site_pos'][0]*0.6
        self.assertEqual(round(sum_sites,1), 0.2)

    def test__contains__T(self):
        self.assertTrue('site_pos' in self.params)

    def test__len__T(self):
        self.assertEqual(len(self.params['site_pos']), 3)

    def test__str__T(self):
        self.assertEqual(str(self.params['site_pos']), '[0.3, 0.5, 0.7]')

    def test__repr__T(self):
        self.assertEqual(repr(self.params['site_pos']), '[0.3, 0.5, 0.7]')

    def test__reversed__T(self):
        self.assertIsInstance(reversed(self.params['site_pos']), collections.abc.Iterable)

    def test_count_T(self):
        self.params.update(num_sites=4, num_chrom=[19,1,2,4],site_pos=[0.4,0.2,0.3,0.1])
        self.assertEqual(self.params['num_chrom'].count(2), 1 )

    def test_index_T(self):
        self.params.update(num_sites=4, num_chrom=[19,1,2,4],site_pos=[0.4,0.2,0.3,0.1])
        self.assertEqual(self.params['site_pos'].index(0.2), 1)

    def test_index_E(self):
        self.params.update(num_sites=4, num_chrom=[19,1,2,4],site_pos=[0.4,0.2,0.3,0.1])
        with self.assertRaises(ValueError):
            self.params['num_chrom'].index(10)

class ParamMatrix_test(unittest.TestCase):

    def setUp(self):
        self.coal= egglib.coalesce.Simulator(6, num_chrom=[20, 20, 20, 20, 1, 10],num_sites=1000, num_alleles=4,
                                                 theta=1.5, migr=0.0, rand_start=True)
        self.params = self.coal.params
        self.params['migr_matrix'] = [[None, 0.75, 0.65, 0.15, 0.00, 0.45],
                           [0.95, None, 0.65, 0.85, 0.00, 0.75],
                           [0.55, 0.75, None, 0.75, 0.00, 0.75],
                           [0.5, 0.56, 0.55, None, 0.00, 0.15],
                           [0.00, 0.00, 0.00, 0.00, None, 0.00],
                           [0.05, 0.78, 0.72, 0.77, 0.00, None]]

    def tearDown(self):
        del self.coal

    def test_ParamMatrix_T(self):
        self.assertIsInstance(self.params['migr_matrix'],  egglib.coalesce._param_helpers.ParamMatrix)

    def test_count_T(self):
        self.assertEqual(self.params['migr_matrix'].count(2.0),0)

    def test_index_T(self):
            self.assertEqual(self.params['migr_matrix'].index(0.77 ),(5, 3))

    def test_index_E(self):
        with self.assertRaises(ValueError):
            self.params['migr_matrix'].index(1)

    def test_get_value_T(self):
        m1=self.params['migr_matrix']
        coal= egglib.coalesce.Simulator(6, num_chrom=[20, 20, 20, 20, 1, 10],num_sites=1000, num_alleles=4,
                                                 theta=1.5, migr=0.0, rand_start=True)
        coal.params['migr_matrix'].get_values(self.params['migr_matrix'])
        m2=coal.params['migr_matrix']
        self.assertNotEqual(m1,m2)

    def test_get_value_E(self):
        coal= egglib.coalesce.Simulator(num_pop=4)
        params = coal.params
        params.update(num_sites=4)
        params.update(num_chrom=[19,1,2,4],site_pos=[0.4,0.2,0.3,0.1], num_alleles=2)
        params['migr_matrix'][0,1] = 0.5
        params['migr_matrix'][1,0] = 1.0 

        with self.assertRaises(ValueError):
            params['migr_matrix'].get_values(self.params['migr_matrix'])

        with self.assertRaises(ValueError):
            self.params['migr_matrix'].get_values(((None, 1.0, 1.0),(0.5, None, 0.5),(0.5,1.0,None)))
        
        with self.assertRaises(ValueError):
            self.params['migr_matrix'].get_values([[None, 1.5],[1.0, None,0.5]])
        

    def test__getitem__T(self):
        self.assertEqual(self.params['migr_matrix'][3,2],0.55)

    def test__setitem__T(self):
        self.params['migr_matrix'][3,2]=0.4
        self.assertEqual(self.params['migr_matrix'][3,2],0.4)

    def test__iter__T(self):
        self.assertIsInstance(self.params['migr_matrix'], collections.abc.Iterable)

    def test__add__T(self):
        sum_mirg=self.params['migr_matrix'][3,2]+self.params['migr_matrix'][3,1]
        self.assertEqual(sum_mirg,  1.11)

    def test__radd__T(self):
        sum_mirg=self.params['migr_matrix'][3,2]+0.2
        self.assertEqual(round(sum_mirg,1), 0.8)

    def test__mul__T(self):
        mul_mrg=self.params['migr_matrix'][3,2]*self.params['migr_matrix'][3,1]
        self.assertEqual(round(mul_mrg,1), 0.3)

    def test__rmul__T(self):
        mul_mrg=self.params['migr_matrix'][3,2]*0.6
        self.assertEqual(round(mul_mrg,1), 0.3)

    def test__contains__T(self):
        self.assertTrue(0.85 in self.params['migr_matrix'])
        self.assertFalse(0.18 in self.params['migr_matrix'])

    def test__len__T(self):
        self.assertEqual(len(self.params['migr_matrix']), 6)

    def test__str__T(self):
        self.assertEqual(str(self.params['migr_matrix']), '[[None, 0.75, 0.65, 0.15, 0.0, 0.45], [0.95, None, 0.65, 0.85, 0.0, 0.75], [0.55, 0.75, None, 0.75, 0.0, 0.75], [0.5, 0.56, 0.55, None, 0.0, 0.15], [0.0, 0.0, 0.0, 0.0, None, 0.0], [0.05, 0.78, 0.72, 0.77, 0.0, None]]')

    def test__repr__T(self):
        self.assertEqual(repr(self.params['migr_matrix']), '<matrix of 6*6 values>')

    def test__reversed__T(self):
        self.assertIsInstance(reversed(self.params['migr_matrix']), collections.abc.Iterable)

class EventsList_test(unittest.TestCase):

    def setUp(self):
        self.coal = egglib.coalesce.Simulator(3, num_chrom=(20, 2, 2), migr=0, theta=0.0)
        self.params=self.coal.params

        self.params['events'].add(cat='size', T=0.8, N=1, idx=0)
        self.params['events'].add(cat='migr', T=0.8, M=1)
        self.params['events'].add(cat='pair_migr', T=0.8, M=1, src=1, dst=0)
        self.params['events'].add(cat='growth', T=0.8, G=1, idx=0)
        self.params['events'].add(cat='selfing', T=0.8, s=1, idx=0)
        self.params['events'].add(cat='recombination', T=0.8, R=1)
        self.params['events'].add(cat='admixture', T=0.8, proba=1, dst=0, src=1)
        self.params['events'].add(cat='sample', T=0.8, idx=1, label='0', num_chrom=2, num_indiv=3)

    def tearDown(self):
        del self.coal

    def test_EventsList_T(self):
        self.assertIsInstance(self.params['events'], egglib.coalesce._param_helpers.EventList)

    def test_replace_T(self):
        coal = egglib.coalesce.Simulator(3, num_chrom=(20, 2, 2), migr=0, theta=0.0)
        coal.params['events'].add(cat='bottleneck', T=0.5, idx=0, S=1.0)
        coal.params['events'].add(cat='merge', T=0.8, src=1, dst=0)
        ncat0=len(self.params['events'])
        self.params['events'].replace(coal.params['events'])
        ncat1=len(self.params['events'])
        self.assertTrue(ncat0>ncat1)

    def test_clear_T(self):
        ncat0=len(self.params['events'])
        self.params['events'].clear()
        ncat1=len(self.params['events'])
        self.assertEqual(len(self.params['events']),0)
        self.assertTrue(ncat0>ncat1)

    def test__len__T(self):
        self.assertEqual(len(self.params['events']), 8)

    def test__iter__T(self):
        self.assertIsInstance(self.params['events'], collections.abc.Iterable)

    def test__str__T(self):
        string = str(self.params['events'])
        self.assertIsInstance(string, str)
        mo = re.match('\[<(.+?)>, <(.+?)>, <(.+?)>, <(.+?)>, <(.+?)>, <(.+?)>, <(.+?)>, <(.+?)>\]', string)
        self.assertIsNotNone(mo)
        groups = mo.groups()
        groups = [dict([i.split('=') for i in g.split(';')]) for g in groups]
        self.assertDictEqual(groups[0], {'event_index': '0', 'T': '0.8', 'idx': '0', 'N': '1', 'cat': 'size'})
        self.assertDictEqual(groups[1], {'event_index': '1', 'M': '1', 'T': '0.8', 'cat': 'migr'})
        self.assertDictEqual(groups[2], {'event_index': '2', 'src': '1', 'dst': '0', 'M': '1', 'cat': 'pair_migr', 'T': '0.8'})
        self.assertDictEqual(groups[3], {'event_index': '3', 'G': '1', 'T': '0.8', 'idx': '0', 'cat': 'growth'})
        self.assertDictEqual(groups[4], {'event_index': '4', 's': '1', 'T': '0.8', 'idx': '0', 'cat': 'selfing'})
        self.assertDictEqual(groups[5], {'event_index': '5', 'R': '1', 'T': '0.8', 'cat': 'recombination'})
        self.assertDictEqual(groups[6], {'event_index': '6', 'src': '1', 'dst': '0', 'cat': 'admixture', 'T': '0.8', 'proba': '1'})
        self.assertDictEqual(groups[7], {'event_index': '7', 'idx': '1', 'cat': 'sample', 'num_indiv': '3', 'T': '0.8', 'label': '0', 'num_chrom': '2'})

    def test__repr__T(self):
        string = repr(self.params['events'])
        self.assertIsInstance(string, str)
        mo = re.match('\[<(.+?)>, <(.+?)>, <(.+?)>, <(.+?)>, <(.+?)>, <(.+?)>, <(.+?)>, <(.+?)>\]', string)
        self.assertIsNotNone(mo)
        groups = mo.groups()
        groups = [dict([i.split('=') for i in g.split(';')]) for g in groups]
        self.assertDictEqual(groups[0], {'event_index': '0', 'T': '0.8', 'idx': '0', 'N': '1', 'cat': 'size'})
        self.assertDictEqual(groups[1], {'event_index': '1', 'M': '1', 'T': '0.8', 'cat': 'migr'})
        self.assertDictEqual(groups[2], {'event_index': '2', 'src': '1', 'dst': '0', 'M': '1', 'cat': 'pair_migr', 'T': '0.8'})
        self.assertDictEqual(groups[3], {'event_index': '3', 'G': '1', 'T': '0.8', 'idx': '0', 'cat': 'growth'})
        self.assertDictEqual(groups[4], {'event_index': '4', 's': '1', 'T': '0.8', 'idx': '0', 'cat': 'selfing'})
        self.assertDictEqual(groups[5], {'event_index': '5', 'R': '1', 'T': '0.8', 'cat': 'recombination'})
        self.assertDictEqual(groups[6], {'event_index': '6', 'src': '1', 'dst': '0', 'cat': 'admixture', 'T': '0.8', 'proba': '1'})
        self.assertDictEqual(groups[7], {'event_index': '7', 'idx': '1', 'cat': 'sample', 'num_indiv': '3', 'T': '0.8', 'label': '0', 'num_chrom': '2'})

    def test__getitem__T(self):
        self.assertEqual(self.params['events'][2],  {'src': 1, 'dst': 0, 'M': 1, 'T': 0.8, 'cat': 'pair_migr'})

    def test_add_T(self):
        ncat0=len(self.params['events'])
        self.params['events'].add(cat='bottleneck', T=0.5, idx=0, S=1.0)
        ncat1= len(self.params['events'])
        self.assertTrue(ncat0 < ncat1)

    def test_add_E(self):
        with self.assertRaises(ValueError):
            self.params['events'].add(cat='FAIL', T=0.5, idx=0, S=1.0)
        with self.assertRaises(ValueError):
            self.params['events'].add(cat='bottleneck', T=0.5, idx=0)
        with self.assertRaises(ValueError):
            self.params.add_event('merge', T=1.0, src=-150, dst=3)
            self.params.add_event('merge', T=1.0, src=2, dst=10)
            self.params.add_event('merge', T=1.0, src=2, dst=2)

    def test_update_T(self):
        self.params['events'].add(cat='bottleneck', T=0.5, idx=0, S=1.0)
        self.params['events'].add(cat='merge', T=0.8, src=1, dst=0)
        btlnk0=str(self.params['events'][8])
        self.params['events'].update(8, T=1, idx=2, S=0.5)
        btlnk1=str(self.params['events'][8])
        self.assertNotEqual(btlnk0,btlnk1)

        mrg0=str(self.params['events'][9])
        self.params['events'].update(9, T=1, idx=3, S=1)
        mrg1=str(self.params['events'][9])
        self.assertNotEqual(mrg0,mrg1)

        self.params['events'].add(cat='sample', T=1.44, num_chrom=0, num_indiv=10, label='NorwegianBlue', idx=0)
        self.params['events'].update(10, T=2.88)
        self.assertEqual(self.params['events'][10]['T'], 2.88)

        self.params['events'].add(cat='bottleneck', T=0.72, idx=0, S=0.1)
        self.params['events'].update(11, S=0.247)
        self.assertEqual(self.params['events'][11]['S'], 0.247)

    def test_update_E(self):
        self.params['events'].add(cat='bottleneck', T=0.5, idx=0, S=1.0)
        self.params['events'].add(cat='merge', T=0.8, src=1, dst=0)

        with self.assertRaises(IndexError): #invalid event index
            self.params['events'].update(12, T=1, src=-3, dst=0)
        with self.assertRaises(ValueError): #src: population index out of range <0
            self.params['events'].update(9, T=1, src=-3, dst=0) 
        with self.assertRaises(ValueError): #src: population index out of range >self._npop
            self.params['events'].update(9, T=1, src=4, dst=0)
        with self.assertRaises(ValueError): #dst: population index out of range <0
            self.params['events'].update(9, T=1, src=3, dst=-10)
        with self.assertRaises(ValueError): #dst: population index out of range >self._npop
            self.params['events'].update(9, T=1, src=3, dst=4)

        with self.assertRaises(ValueError): #T
            self.params['events'].update(8, T=-0.5, idx=0, S=1.0) 
        with self.assertRaises(ValueError): #S
            self.params['events'].update(8, T=0.5, idx=0, S=-1.0) 
        with self.assertRaises(ValueError): #idx
            self.params['events'].update(8, T=0.5, idx=-10, S=1.0) 
        with self.assertRaises(ValueError): #N
            self.params['events'].update(0, T=1, N=-1, idx=0)
        with self.assertRaises(ValueError): #M
            self.params['events'].update(1, T=0.8, M=-0.5)
        with self.assertRaises(ValueError):#src
            self.params['events'].update(2, T=1, src=-3, dst=0) 
        with self.assertRaises(ValueError):#s
            self.params['events'].update(4, T=0.8, s=-1, idx=0)
        with self.assertRaises(ValueError):#R
            self.params['events'].update(5, T=0.8, R=-1)
        with self.assertRaises(ValueError):#proba
            self.params['events'].update(6, T=0.8, proba=-1, dst=0, src=1)
        with self.assertRaises(ValueError): #src
            self.params['events'].update(6, T=0.8, proba=1, dst=0, src=-1)
        with self.assertRaises(ValueError): #dst
            self.params['events'].update(6, T=0.8, proba=1, dst=41, src=1)
        with self.assertRaises(TypeError): #label
            self.params['events'].update(7, T=0.8, idx=1, label=0, num_chrom=2, num_indiv=3)
        with self.assertRaises(ValueError):#num_chrom
            self.params['events'].update(7, T=0.8, idx=1, label='1', num_chrom=-2, num_indiv=3)
        with self.assertRaises(ValueError): #num_indiv
            self.params['events'].update(7, T=0.8, idx=1, label='1', num_chrom=2, num_indiv=-3)



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest, pathlib
path = pathlib.Path(__file__).parent / '..' / 'data'

class ProbaMisoriented_test(unittest.TestCase):
    def test_ProbaMisoriented_T(self):
        PM=egglib.stats.ProbaMisoriented()
        self.assertEqual(str(type(PM)), "<class 'egglib.stats._baudry.ProbaMisoriented'>")

    def test_reset_T(self):
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        PM=egglib.stats.ProbaMisoriented(aln)
        n_sites_a=PM.S
        PM.reset()
        n_sites_b= PM.S
        self.assertTrue(n_sites_a>n_sites_b)

    def test_load_align_T(self):
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        PM=egglib.stats.ProbaMisoriented()
        PM.load_align(aln, egglib.struct_from_labels(aln))
        self.assertTrue(PM.S>0)	

    def test_load_site_T(self):
        PM=egglib.stats.ProbaMisoriented()
        site_null=[[(4,2),(4,3),(3,1)]]
        frq=egglib.Freq()
        frq.from_list(site_null, outgroup=None)
        n_sites_b=PM.S
        PM.load_site(frq)
        n_sites_a=PM.S
        self.assertTrue(n_sites_a>n_sites_b)

    def test_compute_T(self):
        PM=egglib.stats.ProbaMisoriented()
        site=egglib.Site()
        frq=egglib.Freq()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        PM.load_align(aln, egglib.struct_from_labels(aln))
        aln2 = egglib.io.from_fasta(str(path / 'dmi3.fas'), alphabet=egglib.alphabets.Alphabet('char', 'ATGC?', '-MRSYW'), labels=True)
        site.from_align(aln2, 0)
        frq.from_site(site)
        PM.load_site(frq)
        titv1= PM.TiTv
        PM.compute()
        titv2= PM.TiTv
        self.assertAlmostEqual(titv1,4.02298850575, 11)
        self.assertAlmostEqual(titv2,3.97727272727, 11)

    def test_S_T(self):
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        PM=egglib.stats.ProbaMisoriented()
        PM.load_align(aln, egglib.struct_from_labels(aln))
        self.assertEqual(PM.S, 241)

    def test_D_T(self):
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        PM=egglib.stats.ProbaMisoriented()
        PM.load_align(aln, egglib.struct_from_labels(aln))
        self.assertEqual(PM.D, 25)

    def test_TiTv_T(self):
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        PM=egglib.stats.ProbaMisoriented()
        PM.load_align(aln, struct=egglib.struct_from_labels(aln))
        self.assertAlmostEqual(PM.TiTv, 4.02298850575, 11)

    def test_pM_T(self):
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        PM=egglib.stats.ProbaMisoriented()
        PM.load_align(aln, egglib.struct_from_labels(aln))
        self.assertAlmostEqual(PM.pM, 0.141393462231, 12)



"""
    Copyright 2009-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import subprocess, re, os
from .. import eggwrapper as _eggwrapper
from .. import alphabets
from .. import _interface
from ..io import _fasta
from . import _utils

class _Clustal(_utils._App):

    @_utils._protect_run
    def _check_path(self, path, cfg):
        cmd = (path, '--version')
        try:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
            stdout, stderr = p.communicate()
            if len(stderr): return stderr
        except OSError as e:
            return e.strerror
        mo = re.search('(\d+)\.(\d+)\.(\d+)', stdout)
        M = int(mo.group(1))
        m = int(mo.group(2))
        if not mo: return 'cannot read version number from Clustal Omega'
        if M >= 2 or m >= 1: return None
        else: return 'Clustal Omega version 1.1.0 or higher is required'

_app = _Clustal(key='clustal', default='clustalo')
_utils.paths._add(_app)

@_utils._protect_run
def clustal(source, ref=None,
            full=False, full_iter=False, cluster_size=100,
            use_kimura=False, num_iter=1, threads=1, keep_order=False,
            verbose=False):

    """ clustal(source, ref=None, \
                full=False, full_iter=False, cluster_size=100, \
                use_kimura=True,  num_iter=1, threads=1, keep_order=False, \
                verbose=False)

    Multiple sequence alignment using `Clustal Omega <http://www.clustal.org/omega/>`_.

    :param source: a :class:`.Container` or :class:`.Align` containing
        the sequences to align. If a :class:`!Container` is provided,
        sequences are assumed to be unaligned, and, if a :class:`!Align`
        is provided, sequences are assumed to be aligned. The list below
        explains what is done based on the type of *source* and whether
        a value is provided for *ref*:

        * If *source* is a :class:`!Container` and *ref* is ``None``,
          the sequences in *source* are aligned.

        * If *source* is an :class:`!Align` and *ref* is ``None``, a
          hidden Markov model is built from the alignment, then the
          alignment is reset and sequences are realigned.

        * If *source* is an :class:`!Align` and an alignment is
          provided as  *ref*, the two alignments are preserved (their
          columns are left unchanged), and they aligned with respect to
          each other.

        * If *source* is a :class:`!Container` and an alignment is
          provided as *ref*, a hidden Markov model is built from *ref*,
          then *source* is aligned using it, and finally the resulting
          alignment is aligned with respect to *ref* as described for
          the previous case.

        *source* must contain at least two sequences unless it is an
        :class:`!Align` and a value is provided for *ref* (in that case,
        it must contain at least one sequence).

    :param ref: an :class:`!Align` instance providing an external
        alignment. See above for more details. *ref* must contain at
        least one sequence. Sequences must be aligned.

    :param full: use full distance matrix to determine guide tree (the
        default is using the faster and less memory-intensive mBed
        approximation).

    :param full_iter: use full distance matrix to determine guide tree
        during iterations.

    :param cluster_size: size of clusters (as a number of sequences)
        used in the mBed algorithm.

    :param use_kimura: use Kimura correction for estimating
        whole-alignment distance (only available if a protein alignment
        has been provided as *source*).

    :param num_iter: number of iterations allowing to improve the
        quality of the alignment. Must be a number :math:`\geq 1` or a pair
        of numbers :math:`\geq 1`. If the value is a pair of numbers, they
        specify the number of guide tree iterations and hidden Markov
        model iterations, respectively. If a single value is provided,
        iterations couple guide tree and hidden Markov model.

    :param threads: number of threads for parallelization (available for
        parts of the program).

    :param keep_order: return the sequences in the same order as they
        were loaded.

    :param verbose: display Clustal Omega's console output.

    :return: An :class:`.Align` instance containing aligned sequences.

    .. versionchanged:: 3.0.0

        Ported to Clustal Omega and added support for more options.

    .. versionchanged:: 3.1.0

        Support protein sequences.
    """

    # check that program is available
    path = _app.get_path()
    if path is None:
        raise RuntimeError('Clustal Omega program not available -- please configure path')
    command_line = [path, '--infmt=fa', '-o', 'o']

    # mapping for sample reference
    mapping = {}

    # check that source has valid alphabet and write it down
    if source._alphabet == alphabets.DNA: command_line.append('--seqtype=DNA')
    elif source._alphabet == alphabets.protein: command_line.append('--seqtype=Protein')
    else: raise ValueError('alphabet must be DNA or protein')
    _utils._write(source, 'i', mapping)

    # if ref is loaded, write it down
    if ref != None:
        if not isinstance(ref, _interface.Align): raise TypeError('invalid type for `ref` argument')
        if ref._alphabet != source._alphabet: raise ValueError('alphabet of `source` and `ref` must match')
        if len(ref) < 1: raise ValueError('not enough sequences loaded to perform alignment')
        _utils._write(ref, 'p', mapping)
        command_line.append('--profile1=p')

    # if source is a Container, write it as -i argument
    if isinstance(source, _interface.Container):
        if len(source) < 2: raise ValueError('not enough sequences loaded to perform alignment')
        command_line.append('--infile=i')

        # if all sequences have the same length, ensure it is treated as non-aligned sequences
        for sam in source:
            if len(sam.sequence) != source.ls(0): break
        else: command_line.append('--dealign')

    # if an alignment is loaded
    elif isinstance(source, _interface.Align):

        # ensure it is treated as a profile
        command_line.append('--is-profile')

        # if the source is a single Align, write it as -i argument also
        if ref is None:
            if len(source) < 2: raise ValueError('not enough sequences loaded to perform alignment')
            command_line.append('--infile=i')

        # if two Align, write this one as profile2 (2!)
        else:
            if len(source) < 1: raise ValueError('not enough sequences loaded to perform alignment')
            command_line.append('--profile2=i')

    # catch type error
    else: raise TypeError('invalid type for `source` argument')

    # other options
    if verbose: command_line.append('-v')
    if full: command_line.append('--full')
    if full_iter: command_line.append('--full-iter')
    if use_kimura:
        if isinstance(source, _interface.Container):
            raise ValueError('Kimura correction requires aligned sequences')
        if source.alphabet != alphabets.protein:
            raise ValueError('Kimura correction requires protein sequences')
        command_line.append('--use-kimura')
    if cluster_size < 1: raise ValueError('invalid value for `cluster_size` argument')
    command_line.append('--cluster-size={0}'.format(cluster_size))
    if isinstance(num_iter, int):
        if num_iter < 1: raise ValueError('invalid value for `num_iter` argument')
        command_line.append('--iter={0}'.format(num_iter))
    else:
        try:
            n1, n2 = num_iter
            n1 = int(n1)
            n2 = int(n2)
        except (ValueError, TypeError): raise ValueError('invalid value for `num_iter` argument')
        command_line.append('--iter={0}'.format(n1))
        command_line.append('--max-hmm-iterations={0}'.format(n2))
    if threads < 1: raise ValueError('invalid value for `threads` argument')
    command_line.append('--threads={0}'.format(threads))
    if keep_order: command_line.append('--output-order=input-order')
    else: command_line.append('--output-order=tree-order')

    # run the program
    p = subprocess.Popen(command_line, stdin=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        stdout=(None if verbose else subprocess.PIPE), universal_newlines=True) # add universal_newlines to support type `str` as stdin in python3
    stdout, stderr = p.communicate('Y\n')

    # check error
    stderr = stderr.strip()
    if len(stderr):
        raise RuntimeError('error while running Clustal Omega: {0}'.format(stderr))
    if not os.path.isfile('o'):
        raise RuntimeError('unknown error while running Clustal Omega (try verbose mode)')

    # get alignment
    aln = _fasta.from_fasta('o', source.alphabet, labels=False, cls=_interface.Align)

    # set proper names and groups
    for sam in aln:
        ref = mapping[sam.name]
        sam.name = ref.name
        sam.labels = ref.labels

    # return
    return aln



"""
    Copyright 2016-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import numbers
from . import eggwrapper as _eggwrapper
from . import alphabets, _interface

def freq_from_site(site, struct=None):
    """
    Compute allele frequencies from a site.

    :param site: a :class:`.Site` instance.
    :param struct: this argument can be:

        * A :class:`.Structure` instance , allowing to select a subset
          of samples and/or define the structure.
        * A :class:`list` (or compatible) of at least one integer
          providing the sample size (as numbers of samples) of all
          populations, assuming the individuals are organized in the
          corresponding order (all samples of a given population
          grouped together). Samples are grouped in haploid
          individuals.
        * ``None`` (no structure, all samples placed in haploid
          individuals and in a single population).

    :return: A new :class:`.Freq` instance.
    """
    obj = Freq()
    obj.from_site(site, struct)
    return obj

def freq_from_list(ingroup, outgroup, geno_list=None, alphabet=None):
    """
    Import allele frequencies.

    :param ingroup: a list of genotype or allele frequencies (based
        on the value of *geno_list*). There must be one item for each
        allele. If there is a single list, a unique population is
        assumed. To specify multiple populations, the user must pass a
        list of lists (each inner list containing the same number of
        items representing allele frequencies). To specify multiple
        clusters, the user must pass a list of lists of lists of allele
        frequencies. There may also be single-item lists when necessary.
        The frequencies must be null or positive integers. The number of
        frequencies per population is required to be constant for all
        populations (corresponding to the number of alleles or genotypes).
        If *geno_list* is defined, data must be the frequencies of the
        provided genotypes, in the same order. Otherwise, data must be
        allelic frequencies, in the order of increasing allele index (in
        the latter case, data will be loaded as haploid).
    :param outgroup: a list of allele/genotype frequencies for the
        outgroup. The number of alleles or genotypes is also required to
        match. If ``None``, no outgroup samples (equivalent to a list of
        zeros).
    :param geno_list: list of genotypes. Genotypes must be provided as
        tuples or lists. Their length is equal to the ploidy and is
        required to be at least one and constant for all genotypes. For
        haploid data, it is still required to passed allelic values as
        one-item lists (with a ploidy of one). Order
        of alleles within genotypes is significant. If ``None``, data are
        loaded as haploid alleles and the index of alleles is taken as
        allelic value.
    :param alphabet: alphabet to be used to describe alleles. Required
        if *geno_list* is used. By default, use a 

    :return: A new :class:`.Freq` instance.

    .. note::
    
        It is required that there is at least one cluster and one
        population.
    """
    obj = Freq()
    obj.from_list(ingroup, outgroup, geno_list, alphabet)
    return obj

def freq_from_vcf(vcf):
    """
    Import allelic frequencies from a VCF parser. The VCF parser
    must have processed a variant and the variant is required to have
    frequency data available as the AC format field along with the AN
    field. An exception is raised otherwise.

    This function only imports haploid allele frequencies in the ingroup
    (without structure). The first allele is the reference, by
    construction, then all the alternate alleles in the order in which
    they are provided in the VCF file.

    :param vcf: a :class:`.VcfParser` instance containing data. There
        must at least one sample and the AN/AC format fields must be
        available. It is not required to extract variant data manually.
    """
    obj = Freq()
    obj.from_vcf(vcf)
    return obj

class Freq(object):
    """
    Hold allelic and genotypic frequencies for a single site.
    Instances can be created using the three functions :func:`.freq_from_site`,
    :func:`.freq_from_list`, and :func:`.freq_from_vcf`, or using the
    default constructor. After it is created by any way, instances can be
    re-used (which is faster), using their methods :meth:`~.Freq.from_site`,
    :meth:`~.Freq.from_list`, and :meth:`~.Freq.from_vcf`.

    .. _compartments:

    :cvar ingroup: whole ingroup.
    :cvar outgroup: outgroup.
    :cvar cluster: a specific cluster (identified by its index).
    :cvar population: a specific population (identified by its index).

    The variables above specify what subset of data should be
    considered. You should never try to modify them. For clusters and
    populations, an index should be provided as well.

    They can be used as::

        >>> freq.freq_allele(allele_index, cpt=egglib.Freq.ingroup)
        >>> freq.freq_allele(allele_index, cpt=egglib.Freq.population, idx=pop_index)

    or::

        >>> freq.freq_allele(allele_index, cpt=freq.ingroup)
        >>> freq.freq_allele(allele_index, cpt=freq.population, idx=pop_index)

    """
    ingroup = 0
    outgroup = 1
    cluster = 2
    population = 3

    def __init__(self):
        self._obj = _eggwrapper.FreqHolder()
        self._alphabet = None

    def from_site(self, site, struct=None):
        """
        Import frequencies from a site.
        Reset the instance as if it was created using
        :func:`.freq_from_site`. Arguments are identical to this function.
        """
        if struct is None:
            struct = _eggwrapper.StructureHolder()
            struct.mk_dummy_structure(site._obj.get_ns(), 1)
        elif isinstance(struct, _interface.Structure):
            struct = struct._obj
            if struct.get_req() > site._obj.get_ns(): raise ValueError('invalid structure (sample index out of range)')
        else:
            if len(struct) < 1: raise ValueError('there must be at least one population size')
            if sum(struct) != site._obj.get_ns(): raise ValueError('invalid structure (sample size is required to match)')
            struct = _eggwrapper.StructureHolder()
            struct.mk_dummy_structure(struct[0], 1)
            for i in struct[1:]: struct.dummy_add_pop(i)
        self._obj.setup_structure(struct)
        self._obj.process_site(site._obj)
        self._alphabet = site._alphabet

    def from_list(self, ingroup, outgroup, geno_list=None, alphabet=None):
        """ Import frequencies from lists.
        Reset the instance as if it was created using :func:`.freq_from_list`.
        Arguments are identical to this function.
        """

        # reformat ingroup argument
        if all([isinstance(i, numbers.Integral) for i in ingroup]):
            ingroup = [[ingroup]]
        elif all([isinstance(j, numbers.Integral) for i in ingroup for j in i]):
            ingroup = [ingroup]
        else:
            if not all([isinstance(k, numbers.Integral) for i in ingroup for j in i for k in j]):
                raise ValueError('invalid form for the `ingroup` argument')

        # get structure properties
        nc = len(ingroup)
        ns = [sum(pop) for clu in ingroup for pop in clu]
        np = len(ns)
        ni = sum(ns)

        # get number of genotypes
        ng = set(map(len, [j for i in ingroup for j in i]))
        if outgroup is not None: ng.add(len(outgroup))
        if len(ng) != 1: raise ValueError('number of frequencies must be the same for all populations and outgroup')
        ng = ng.pop()
        if ng < 1: raise ValueError('there must be at least one allele')
        if outgroup is None: outgroup = [0] * ng
        no = sum(outgroup)
        if geno_list is None:
            pl = 1
            geno_list = [[i,] for i in range(ng)]
            if alphabet is None:
                self._alphabet = alphabets.positive_infinite
            else:
                self._alphabet = alphabet
                try: [self._alphabet.get_value(i) for i in range(ng)]
                except ValueError:
                    raise ValueError('alphabet does not have enough values')

        else:
            if ng != len(geno_list): raise ValueError('invalid number of genotypes')
            pl = set(map(len, geno_list))
            if len(pl) != 1: raise ValueError('ploidy is not constant in genotypes')
            pl = pl.pop()
            if pl < 1: raise ValueError('ploidy must be at least 1')
            geno_set = set()
            for g in geno_list:
                if g in geno_set: raise ValueError('genotype {0} is repeated'.format(g))
                geno_set.add(g)
            if alphabet is None: raise ValueError('alphabet must be specified')
            self._alphabet = alphabet

        # list with allele index (not code!) instead of allele value from geno_list
        geno_indexes = [[None for j in i] for i in geno_list]

        # setup the instance
        self._obj.setup_raw(nc, np, pl)
        idx = 0
        for i, clu in enumerate(ingroup):
            for j, pop in enumerate(clu):
                self._obj.setup_pop(idx, i, j, sum(pop))
                idx += 1
        self._obj.set_ngeno(ng)
        for i, g in enumerate(geno_list):
            for j, v in enumerate(g):
                v = self._alphabet.get_code(v)
                if v < 0: raise ValueError('missing data found in genotypes')
                self._obj.set_genotype_item(i, j, v)
                geno_indexes[i][j] = self._obj.find_allele(v)

        # load data
        pop_idx = 0
        ing_frq = self._obj.frq_ingroup()
        for clu_idx, clu_data in enumerate(ingroup):
            clu_frq = self._obj.frq_cluster(clu_idx)
            for pop_data in clu_data:
                pop_frq = self._obj.frq_population(pop_idx)
                for i, n in enumerate(pop_data):
                    ing_frq.incr_genotype(i, n)
                    clu_frq.incr_genotype(i, n)
                    pop_frq.incr_genotype(i, n)
                    for j in geno_indexes[i]:
                        ing_frq.incr_allele(j, n)
                        clu_frq.incr_allele(j, n)
                        pop_frq.incr_allele(j, n)
                pop_idx += 1

        otg_frq = self._obj.frq_outgroup()
        for i, n in enumerate(outgroup):
            otg_frq.incr_genotype(i, n)
            for j in geno_indexes[i]:
                otg_frq.incr_allele(j, n)

        # load heterozygote genotypes
        if pl > 1:
            for i, g in enumerate(geno_indexes):
                g = set(g)
                if len(g) > 1:
                    for j in g:
                        self._obj.frq_ingroup().tell_het(i, j)
                        self._obj.frq_outgroup().tell_het(i, j)
                        for k in range(self._obj.num_clusters()):
                            self._obj.frq_cluster(k).tell_het(i, j)
                        for k in range(self._obj.num_populations()):
                            self._obj.frq_population(k).tell_het(i, j)

    def from_vcf(self, vcf):
        """ Import frequencies from a VCF file.
        Reset the instance as if it was created using :func:`.freq_from_vcf`.
        Argument is identical to this function.
        """
        if vcf._parser.has_data() == False: raise ValueError('data must have been read from VCF parser')
        if vcf._parser.has_AC() == False or vcf._parser.has_AN() == False: raise ValueError('VCF data must have AC and AN info fields')
        self._obj.process_vcf(vcf._parser)
        self._alphabet = alphabets.positive_infinite

    @property
    def ploidy(self):
        """ Ploidy. """
        return self._obj.ploidy()

    @property
    def num_alleles(self):
        """ Number of alleles in the whole site. """
        return self._obj.num_alleles()

    @property
    def num_genotypes(self):
        """ Number of genotypes in the whole site. """
        return self._obj.num_genotypes()

    @property
    def num_clusters(self):
        """ Number of clusters. """
        return self._obj.num_clusters()

    @property
    def num_populations(self):
        """ Number of populations. """
        return self._obj.num_populations()

    def allele(self, idx):
        """Get an allele."""
        if idx<0 or idx>=self._obj.num_alleles():
            raise IndexError('invalid allele index')
        return self._alphabet.get_value(self._obj.allele(idx))

    def genotype(self, idx):
        """ Get a genotype, as a tuple of alleles. """
        if idx<0 or idx>=self._obj.num_genotypes():
            raise IndexError('invalid genotype index')
        return tuple(self._alphabet.get_value(self._obj.genotype_item(idx, i)) for i in range(self._obj.ploidy()))

    def _getter(self, cpt, idx):
        if cpt == self.ingroup: return self._obj.frq_ingroup()
        elif cpt == self.outgroup: return self._obj.frq_outgroup()
        elif cpt == self.cluster:
            if idx is None:
                raise ValueError('cluster index is required')
            if idx<0 or idx>= self._obj.num_clusters():
                raise IndexError('invalid cluster index')
            return self._obj.frq_cluster(idx)
        elif cpt == self.population:
            if idx is None:
                raise ValueError('population index is required')
            if idx<0 or idx>= self._obj.num_populations():
                raise IndexError('invalid population index')
            return self._obj.frq_population(idx)
        else: raise ValueError('invalid compartment identifier')

    def freq_allele(self, allele, cpt=ingroup, idx=None):
        """
        Get the frequency of an allele.

        :param allele: allele index.
        :param cpt: :ref:`compartment <compartments>` identifier.
        :param idx: compartment index (required for clusters and
            populations, ignored otherwise).
        """
        if allele<0 or allele>=self._obj.num_alleles(): raise IndexError('invalid allele index')
        return self._getter(cpt, idx).frq_all(allele)

    def freq_genotype(self, genotype, cpt=ingroup, idx=None):
        """
        Get the frequency of an genotype.

        :param genotype: genotype index.
        :param cpt: :ref:`compartment <compartments>` identifier.
        :param idx: compartment index (required for clusters and
            populations, ignored otherwise).
        """
        if genotype<0 or genotype>=self._obj.num_genotypes(): raise IndexError('invalid genotype index')
        return self._getter(cpt, idx).frq_gen(genotype)

    def nieff(self, cpt=ingroup, idx=None):
        """
        Get the number of individuals within a given compartment. In
        the haploid case, this method is identical to :meth:`.nseff`.

        :param cpt: :ref:`compartment <compartments>` identifier.
        :param idx: compartment index (required for clusters and
            populations, ignored otherwise).
        """
        return self._getter(cpt, idx).nieff()

    def nseff(self, cpt=ingroup, idx=None):
        """
        Get the number of samples within a given compartment.

        :param cpt: :ref:`compartment <compartments>` identifier.
        :param idx: compartment index (required for clusters and
            populations, ignored otherwise).
        """
        return self._getter(cpt, idx).nseff()



"""
    Copyright 2023 Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from .. import __version__
from ._utils import paths
import click

#### command for editing paths from the console ########################

@click.command(no_args_is_help=True)
@click.version_option(__version__, message='EggLib %(version)s')
@click.option('-a', '--autodetect', is_flag=True, show_default=False, default=False,
    help='Auto-configure application paths based on default command names.')
@click.option('-p', '--path', nargs=2, multiple=True, help='set')
@click.option('-d', '--display', is_flag=True, show_default=False, default=False,
    help='Show values of applications paths (after any setting or detection if requested)')
@click.option('--save/--no-save', '-s/-n', default=False, show_default=True,
    help='Save application paths for future imports of EggLib')
def paths_edit(autodetect, path, display, save):
    """
    Edit external application paths for the current EggLib installation.
    """

    if autodetect:
        paths.autodetect(True)

    for k, v in path:
        paths[k] = v

    if display:
        print('Current values of application paths:')
        for p in paths:
            if paths[p] is None: click.echo(f'    {p}:')
            else: click.echo(f'    {p}: `{paths[p]}\'')

    if save:
        paths.save()
        print(f'Application paths saved to {paths.fname}')
